# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import sys
import os
import threading
from os import path
import tkinter as tk
from tkinter import ttk, filedialog
import numpy as np
from main_ballistics import compute_ballistics

PROGRAM_VERSION = "1.0.2"

# Initial values of each parameters
initial_values = {
    "gravity_m_per_s2": 9.8,
    "fluid_type": "air",
    "temprature_celsius": 25.0,
    "pressure_pa": 101325.0,
    "relative_humidity_percent": 60.0,
    "flow_velocity_m_per_s": "0.0, 0.0, 0.0",
    "mass_g": 0.25,
    "diameter_m": 0.00595,
    "initial_time_s": 0.0,
    "initial_position_m": "0.0, 0.0, 1.0",
    "initial_velocity_m_per_s": "84.85281374, 0.0, 0.0",
    "initial_rotation_rot_per_s": "0.0, -300.0, 0.0",
    "analysis_type": "approx",
    "output_time_step_s": 0.025,
    "num_of_output": 401,
    "output_folder_path": path.abspath("./"),
    "is_output_all_figure": False,
}

# Description for each parameters
param_description = {
    "gravity_m_per_s2": ("General", "Gravity (m/s^2)"),
    "fluid_type": ("Fluid", "Fluid type"),
    "temprature_celsius": ("Fluid", "Temperature (°C)"),
    "pressure_pa": ("Fluid", "Pressure (Pa)"),
    "relative_humidity_percent": ("Fluid", "Humidity (%)"),
    "flow_velocity_m_per_s": ("Fluid", "Flow velocity (m/s)"),
    "mass_g": ("BB", "Mass (g)"),
    "diameter_m": ("BB", "Diameter (m)"),
    "initial_time_s": ("Initial condition", "Time (s)"),
    "initial_position_m": ("Initial condition", "Position (x,y,z) (m)"),
    "initial_velocity_m_per_s": ("Initial condition", "Velocity (x,y,z) (m/s)"),
    "initial_rotation_rot_per_s": ("Initial condition", "Rotation (x,y,z) (rot/s)"),
    "analysis_type": ("Numerics", "Solving EOM"),
    "output_time_step_s": ("Numerics", "Output time step"),
    "num_of_output": ("Numerics", "Num. of time step limit"),
    "output_folder_path": ("Output", " "),
    "is_output_all_figure": ("Output", "Output all figures"),
}


# List of options
options_analysis_type = ["exact", "approx", "approx_w_solved", "no_rotation_decay"]
options_fluid_type = ["air"]


# Dictionary to save input widget
entry_widgets = {}


def calculate():
    # Save input parameters
    with open(os.path.join(entry_widgets["output_folder_path"].get(), "input_parameters.txt"), "w") as f:
        f.write(f"PROGRAM_VERSION,{PROGRAM_VERSION}\n")
        for name, entry in entry_widgets.items():
            f.write(f"{name},{entry.get()}\n")

    # Get each parameters from each entry widget
    kwargs = {}
    for name, entry in entry_widgets.items():
        value = entry.get()
        if name in {"initial_position_m", "initial_velocity_m_per_s", "flow_velocity_m_per_s"}:
            # Convert string to ndarray
            value = np.array(list(map(float, value.split(","))))
            kwargs[name] = value
        elif name in {"initial_rotation_rot_per_s"}:
            # Convert string to ndarray and rot/s to rad/s
            value = np.array(list(map(float, value.split(","))))
            value = value * 2.0 * np.pi
            kwargs["initial_rotation_rad_per_s"] = value
        elif name in {"fluid_type", "analysis_type"}:
            # Convert string
            value = str(value)
            kwargs[name] = value
        elif name in {"mass_g"}:
            # Convert g to kg
            value = float(value) * 0.001
            kwargs["mass_kg"] = value
        elif name in {"diameter_m"}:
            # Convert diameter to radius
            value = float(value) * 0.5
            kwargs["radius_m"] = value
        elif name in {"num_of_output"}:
            # Convert to int
            value = int(value)
            kwargs[name] = value
        elif name in {"output_folder_path"}:
            # Convert to str
            value = str(value)
            kwargs[name] = value
        elif name in {"is_output_all_figure"}:
            # Convert to str
            value = bool(value)
            kwargs[name] = value
        else:
            # Convert to float
            value = float(value)
            kwargs[name] = value

    # Call the routine using obtained kwargs
    label_result.config(text="Status: Calculating...")

    def thread_func():
        try:
            time_sequence, sol = compute_ballistics(**kwargs)
            text_box.insert(tk.END, "Simple Results\n")
            text_box.insert(tk.END, f"  at t={time_sequence[0]:.2f}: x(m)={sol[0,0]:.2f}, z(m)={sol[2,0]:.2f},\n")
            text_box.insert(tk.END, f"  at t={time_sequence[-1]:.2f}: x(m)={sol[0,-1]:.2f}, z(m)={sol[2,-1]:.2f},\n")
            text_box.see(tk.END)
            label_result.config(text="Status: Successful")
        except ValueError as e:
            label_result.config(text=f"Status: Error {e}")

    threading.Thread(target=thread_func).start()


# 　# Folder selection function
def select_folder():
    current_path = folder_path.get()  # Store the current path
    new_path = filedialog.askdirectory()
    if new_path:  # Only update if new_path is not empty
        folder_path.set(new_path)
    else:  # If user cancels the dialog, restore the previous path
        folder_path.set(current_path)


def on_closing():
    root.destroy()
    sys.exit(0)  # End Python process


if __name__ == "__main__":
    root = tk.Tk()

    if getattr(sys, "frozen", False):
        application_path = sys._MEIPASS
    else:
        application_path = os.path.dirname(os.path.abspath(__file__))

    icon_path = os.path.join(application_path, "app.ico")
    root.iconbitmap(icon_path)

    root.title("Airsoft Ballistics Calculator Support GUI")

    current_category = None
    row_idx = 0
    for param_name, initial_value in initial_values.items():
        category, description = param_description[param_name]
        if category != current_category:
            # tk.Label(root, text=category, font=("Arial", 12)).grid(row=row_idx, column=0, columnspan=2)
            tk.Label(root, text=category, anchor="w", font=("Arial", 12)).grid(row=row_idx, column=0, sticky="w", columnspan=2)
            row_idx += 1
            current_category = category
        # tk.Label(root, text=description).grid(row=row_idx, column=0)
        tk.Label(root, text=description, anchor="e").grid(row=row_idx, column=0, sticky="e")
        if param_name == "analysis_type":
            # Generate OptionMenu widget when analysis_type
            # selected_option = tk.StringVar(root)
            # selected_option.set(initial_value)
            # option_menu = tk.OptionMenu(root, selected_option, *options_analysis_type)
            # option_menu.grid(row=row_idx, column=1)
            # entry_widgets[param_name] = selected_option  # Save StringVar

            # Generate ttk.Combobox widget when analysis_type
            combo = ttk.Combobox(root, values=options_analysis_type, width=17)
            combo.set(initial_value)
            combo.grid(row=row_idx, column=1)
            entry_widgets[param_name] = combo  # Save Combobox
        elif param_name == "fluid_type":
            combo = ttk.Combobox(root, values=options_fluid_type, width=17)
            combo.set(initial_value)
            combo.grid(row=row_idx, column=1)
            entry_widgets[param_name] = combo  # Save Combobox
        elif param_name == "output_folder_path":
            folder_path = tk.StringVar()
            folder_path.set(initial_value)  # Set initial folder path

            folder_path = tk.StringVar()
            folder_path.set(initial_value)  # Set initial folder path
            # tk.Button(root, text="Select Folder", command=lambda: folder_path.set(filedialog.askdirectory())).grid(row=row_idx, column=0, sticky="n")
            tk.Button(root, text="Select Folder", command=select_folder).grid(row=row_idx, column=0, sticky="n")
            tk.Label(root, textvariable=folder_path, height=5, wraplength=150, anchor="nw", justify=tk.LEFT).grid(
                row=row_idx, column=1
            )  # Display selected folder path with specific height and wraplength
            entry_widgets[param_name] = folder_path  # Save StringVar
        elif param_name == "is_output_all_figure":
            is_enabled = tk.BooleanVar()
            is_enabled.set(initial_value)  # Set initial checkbox value
            check_button = tk.Checkbutton(root, variable=is_enabled)
            check_button.grid(row=row_idx, column=1)
            entry_widgets[param_name] = is_enabled  # Save BooleanVar
        else:
            entry = tk.Entry(root)
            entry.insert(0, initial_value)
            entry.grid(row=row_idx, column=1)
            entry_widgets[param_name] = entry

        row_idx += 1

    tk.Button(root, text="Calculate", command=calculate).grid(row=row_idx, column=1)

    label_result = tk.Label(root, text="Status: ")
    label_result.grid(row=row_idx + 1, column=0, columnspan=2)

    # Generate text_box
    text_box = tk.Text(root, width=45, height=15)
    text_box.grid(row=row_idx + 2, column=0, columnspan=2)

    label_version = tk.Label(
        root,
        text=f"Version: {PROGRAM_VERSION}",
        anchor="e",
    )
    label_version.grid(row=row_idx + 3, column=1, sticky="e")

    root.protocol("WM_DELETE_WINDOW", on_closing)

    root.mainloop()
