# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import scipy as sp
from numpy.typing import NDArray
from class_def.environment import Environment
from class_def.particle import Particle
from physical_quantities.fluid_quantities import density_kg_per_m3, reynolds_number, mach_number, drag_coefficients_sphere, skin_friction_coefficients


def eom_bb_bullet_approx_w_solved(t: np.float64, x: NDArray[np.float64], p: Particle, env: Environment) -> NDArray[np.float64]:
    x_rhs = np.empty(x.shape[0], dtype=np.float64)

    # Physical parameters
    m = p.mass_kg  # mass of BB bullet
    R = p.radius_m  # radius of BB bullet
    L = p.characteristic_length_m  # characteristic_length_m

    ftype = env.fluid_type
    P = env.pressure_pa  # pressure in Pa
    T = env.temprature_celsius  # temprature in celsius
    g = env.gravity_m_per_s2  # gravitational acceleration
    phi = env.relative_humidity_percent  # relative humidity in percent
    u = env.flow_velocity_m_per_s  # flow velocity in m/s

    # r = x[0:3]
    v = x[3:6]
    V = v - u
    V_norm = np.linalg.norm(V)

    rho = density_kg_per_m3(T, P, phi, ftype)
    Re = reynolds_number(V_norm, L, T, P, phi, ftype)
    Ma = mach_number(V_norm, T, P, phi, ftype)
    Cd = drag_coefficients_sphere(Re, Ma, ftype)
    Cl = 0.12
    Cf = skin_friction_coefficients(Re, ftype)

    u_v_norm = np.linalg.norm(u - v)
    u_v0_norm = np.linalg.norm(u - p.initial_velocity_m_per_s)

    t0 = p.initial_time_s
    w0_norm = np.linalg.norm(p.initial_rotation_rad_per_s)

    # Force
    # - Gravity force
    Fg = np.array([0.0, 0.0, -m * g], dtype=np.float64)
    # - Drag force
    Fd = -0.5 * Cd * rho * np.pi * R * R * V_norm * V
    # - Magnus force
    if w0_norm >= 1e-30:
        w_approx_norm = rotation_decaying_approx(t, R, m, u_v_norm, u_v0_norm, t0, w0_norm, rho, Cf)
        w = w_approx_norm * (p.initial_rotation_rad_per_s / w0_norm)
        V_cross_w = np.cross(V, w)
        Fl = -Cl * ((4.0 * np.pi * R**3) / 3.0) * 2.0 * rho * V_cross_w
    else:
        Fl = np.zeros(3, dtype=np.float64)

    x_rhs[0:3] = x[3:6]
    x_rhs[3:6] = (Fg + Fd + Fl) / m

    return x_rhs


def integral_in_moment_of_force(v: float, w: float, u: float, R: float, component: str):
    theta_a_limits = 0.0
    theta_b_limits = np.pi
    phi_a_limits = 0.0
    phi_b_limits = 2.0 * np.pi

    result, error = sp.integrate.dblquad(
        integrand,
        theta_a_limits,
        theta_b_limits,
        phi_a_limits,
        phi_b_limits,
        args=(
            v,
            w,
            u,
            R,
            component,
        ),
        epsabs=1e-8,
        epsrel=1e-8,
    )

    # print(result, error)
    return result


def integrand(theta, phi, v: NDArray[np.float64], w: NDArray[np.float64], u: NDArray[np.float64], R: float, component: str) -> np.float64:
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    sin_p = np.sin(phi)
    cos_p = np.cos(phi)
    unit_vector_radius = np.array([sin_t * cos_p, sin_t * sin_p, cos_t])
    U = u - np.cross(w, R * unit_vector_radius) - v
    U_norm = np.linalg.norm(U)
    s = U_norm * (np.cross(unit_vector_radius, U))
    if component == "x":
        result = s[0]
    elif component == "y":
        result = s[1]
    elif component == "z":
        result = s[2]
    else:
        raise ValueError(f"Unexpected component in integrand : component={component}")

    return result


def rotation_decaying_approx(
    t: np.float64,
    R: np.float64,
    m: np.float64,
    u_v_norm: np.float64,
    u_v0_norm: np.float64,
    t0: np.float64,
    w0_norm: np.float64,
    rho: np.float64,
    Cf: np.float64,
) -> np.float64:
    const = 0.5  # const is adapted value between [0, 1] (experimentally 0.5)

    Inertia = 0.4 * m * R * R  # Inertia of sphere

    tmp = rho * R * R * (8.0 * np.pi * R * R / 3.0) / (2.0 * Inertia)
    gamma = Cf * tmp * u_v_norm
    gamma0 = Cf * tmp * u_v0_norm
    integration_const = np.arcsinh(u_v0_norm / (const * R * w0_norm))
    # print(gamma, gamma0, gamma * t + integration_const)
    heuristic_coeff = 1.0
    w_norm = (u_v_norm / (const * R)) * (1.0 / np.sinh(heuristic_coeff * (gamma * t - gamma0 * t0) + integration_const))
    return w_norm


def rotation_decaying_approx_wrapper(t: np.float64, v: NDArray[np.float64], p: Particle, env: Environment) -> np.float64:
    ftype = env.fluid_type
    P = env.pressure_pa  # pressure in Pa
    T = env.temprature_celsius  # temprature in celsius
    phi = env.relative_humidity_percent  # relative humidity in percent

    u = env.flow_velocity_m_per_s
    v0 = p.initial_velocity_m_per_s
    u_v_norm = np.linalg.norm(u - v)
    u_v0_norm = np.linalg.norm(u - v0)
    V_norm = u_v_norm
    rho = density_kg_per_m3(T, P, phi, ftype)
    Re = reynolds_number(V_norm, p.characteristic_length_m, T, P, phi, ftype)
    Cf = skin_friction_coefficients(Re, ftype)

    R = p.radius_m
    m = p.mass_kg
    t0 = p.initial_time_s
    w0_norm = np.linalg.norm(p.initial_rotation_rad_per_s)

    w_approx_norm = rotation_decaying_approx(t, R, m, u_v_norm, u_v0_norm, t0, w0_norm, rho, Cf)

    return w_approx_norm
