# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from typing import Union, Dict, Tuple

plt.rcParams["font.family"] = "Times New Roman"  # font
plt.rcParams["font.size"] = 7  # character size
plt.rcParams["mathtext.fontset"] = "stix"  # font Math
plt.rcParams["xtick.direction"] = "in"  # x axis in
plt.rcParams["ytick.direction"] = "in"  # y axis in

plt.rcParams["axes.prop_cycle"] = plt.cycler(
    color=["#000000", "#ff0000", "#0000ff", "#228b22", "#ff00ff", "#ffd900", "#00ffff", "#7f00ff", "#ff7f00", "#00ff7f", "#bebebe", "#666666"]
)


# Color line palette from gnuplot
def plot_xy(
    output_image_name: str,
    x_dict: Dict[str, np.ndarray],
    y_dict: Dict[str, np.ndarray],
    show_GUI: bool = False,
    graph_title: str = "x-y graph",
    graph_title_size: int = 7,
    x_label: str = "${\it x}$",
    x_range: Union[tuple, str] = "auto",
    x_major_locator: Union[float, str] = "auto",
    x_minor_locator: Union[float, str] = "auto",
    x_show_grid: bool = True,
    y_label: str = "${\it y}$",
    y_range: Union[tuple, str] = "auto",
    y_major_locator: Union[float, str] = "auto",
    y_minor_locator: Union[float, str] = "auto",
    y_show_grid: bool = True,
    aspect_ratio: Union[str, float] = "auto",
    figure_size_tuple: Tuple = (4, 4),
    legend_loc_tuple: Tuple = (0.7, 0.7),
) -> None:
    # Check size
    for k in x_dict.keys():
        if x_dict[k].shape[0] != y_dict[k].shape[0]:
            raise ValueError(f"Unexpected y_dict size : x_dict[{k}].shape[0]={x_dict[k].shape[0]}, y_dict[{k}].shape[0]={y_dict[k].shape[0]}")

    # x = np.linspace(0, 2 * np.pi, 20)  # x=[0,2*pi], N=20 点使用 , Δ x=1/(N-1)
    # y = np.sin(x)  # y=sin(x)
    # z = np.cos(x)  # z=cos(x)

    # Make figure and subplots
    fig, axs = plt.subplots(1, 1, figsize=figure_size_tuple)  # Generate figure and axes objects (row, column)

    # Plot data settings
    for k in y_dict.keys():
        axs.plot(x_dict[k], y_dict[k], ".-", ms=3, label=k)  # set graph x,y with circle and line

    axs.legend(loc=legend_loc_tuple)  # Add legend to graph by relative coordinate, loc=(0-1,0-1)
    # Graph title
    axs.set_title(graph_title, fontdict={"fontsize": graph_title_size})

    # x- and y-axis Common
    axs.set_aspect(aspect_ratio, adjustable="box")  # x and y axis aspect
    # ===== x-axis =====
    # x-axis label
    axs.set_xlabel(x_label)  # label set
    # x-axis range
    if isinstance(x_range, tuple):
        axs.set_xlim(x_range[0], x_range[1])  # x-axis describe range = (x_range[0], x_range[1])
    elif isinstance(x_range, str):
        if x_range == "auto":
            axs.set_xlim(auto=True)  # Automatically decide x-axis describe range
        else:
            raise ValueError(f"x_range is a incorrect string, your x_range={x_range}")
    else:
        raise ValueError(f"x_range is neither a tuple nor a string, your x_range={x_range}")

    # x-axis major locator
    if isinstance(x_major_locator, float):
        axs.xaxis.set_major_locator(tck.MultipleLocator(x_major_locator))  # Set x-axis major locator as a value of x_tics_major_locator
    elif isinstance(x_major_locator, str):
        if x_major_locator == "auto":
            axs.xaxis.set_major_locator(tck.AutoLocator())  # Automatically decide x-axis major locator
        else:
            raise ValueError(f"x_major_locator is a incorrect string, your x_major_locator={x_major_locator}")
    else:
        raise ValueError(f"x_major_locator is neither a tuple nor a string, your x_major_locator={x_major_locator}")

    # x-axis minor locator
    if isinstance(x_minor_locator, float):
        axs.xaxis.set_minor_locator(tck.MultipleLocator(x_minor_locator))  # Set x-axis minor locator as a value of x_tics_major_locator
    elif isinstance(x_minor_locator, str):
        if x_minor_locator == "auto":
            axs.xaxis.set_minor_locator(tck.AutoLocator())  # Automatically decide x-axis minor locator
        else:
            raise ValueError(f"x_minor_locator is a incorrect string, your x_minor_locator={x_minor_locator}")
    else:
        raise ValueError(f"x_minor_locator is neither a tuple nor a string, your x_minor_locator={x_minor_locator}")

    # x-axis show the grid or not
    if x_show_grid:
        axs.grid(axis="x", which="major", color="#cccccc", linestyle="--")  # show x-axis grid
        axs.grid(axis="x", which="minor", color="#e7e7e7", linestyle="--")  # show x-axis sub-grid
    else:
        axs.grid(axis="x", which="major", b=False)
        axs.grid(axis="x", which="minor", b=False)

    # ===== y-axis =====
    # y-axis label
    axs.set_ylabel(y_label)  # label set
    # y-axis range
    if isinstance(y_range, tuple):
        axs.set_ylim(y_range[0], y_range[1])  # y-axis describe range = (y_range[0], y_range[1])
    elif isinstance(y_range, str):
        if y_range == "auto":
            axs.set_ylim(auto=True)  # Automatically decide y-axis describe range
        else:
            raise ValueError(f"y_range is a incorrect string, your y_range={y_range}")
    else:
        raise ValueError(f"y_range is neither a tuple nor a string, your y_range={y_range}")

    # y-axis major locator
    if isinstance(y_major_locator, float):
        axs.yaxis.set_major_locator(tck.MultipleLocator(y_major_locator))  # Set y-axis major locator as a value of y_tics_major_locator
    elif isinstance(y_major_locator, str):
        if y_major_locator == "auto":
            axs.yaxis.set_major_locator(tck.AutoLocator())  # Automatically decide y-axis major locator
        else:
            raise ValueError(f"y_major_locator is a incorrect string, your y_major_locator={y_major_locator}")
    else:
        raise ValueError(f"y_major_locator is neither a tuple nor a string, your y_major_locator={y_major_locator}")

    # y-axis minor locator
    if isinstance(y_minor_locator, float):
        axs.yaxis.set_minor_locator(tck.MultipleLocator(y_minor_locator))  # Set y-axis minor locator as a value of y_tics_major_locator
    elif isinstance(y_minor_locator, str):
        if y_minor_locator == "auto":
            axs.yaxis.set_minor_locator(tck.AutoLocator())  # Automatically decide y-axis minor locator
        else:
            raise ValueError(f"y_minor_locator is a incorrect string, your y_minor_locator={y_minor_locator}")
    else:
        raise ValueError(f"y_minor_locator is neither a tuple nor a string, your y_minor_locator={y_minor_locator}")

    # y-axis show the grid or not
    if y_show_grid:
        axs.grid(axis="y", which="major", color="#cccccc", linestyle="--")  # show x-axis grid
        axs.grid(axis="y", which="minor", color="#e7e7e7", linestyle="--")  # show x-axis sub-grid
    else:
        axs.grid(axis="y", which="major", b=False)
        axs.grid(axis="y", which="minor", b=False)

    # Don't plot the left-bottom region because the Bounding Box has negative value
    fig.subplots_adjust(left=0.3, right=0.9, bottom=0.3, top=0.9)
    if show_GUI:
        # Output GUI
        plt.show()  # GUI にプロット出力
    # Save figure
    fig.savefig(output_image_name, dpi=400, bbox_inches="tight")

    plt.close("all")
