# MIT License
# Copyright (c) 2023 sikino
# See LICENSE file in the project root for full license information.

import math
import numpy as np
from typing import Union
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker


def density_kg_per_m3(
    temprature_celsius: Union[float, np.float64],
    pressure_pa: Union[float, np.float64],
    relative_humidity_percent: Union[float, np.float64],
    fluid_type: str,
) -> np.float64:
    #
    # Density of airDensity of air containing vapor at pressure p
    #
    # temprature_celsius < 374.15
    if fluid_type == "air":
        T = temprature_celsius
        T_K = T + 273.15
        H = pressure_pa
        phi = relative_humidity_percent * 0.01

        # Density of dry air
        dryair_density_kg_per_m3 = (1.293 / (1.0 + 0.00367 * T)) * (H / (760.0 * 133.322))

        # Wagner equation for saturation vapor pressure
        #   W. Wagner and A. Pruß:" The IAPWS Formulation 1995 for the Thermodynamic Properties of Ordinary Water Substance for General and Scientific Use",
        #   J. Phys. Chem. Ref. Data, Vol. 31, No. 2, 2002
        Tc_K = 647.096
        theta = 1.0 - T_K / Tc_K
        theta0r5 = math.sqrt(theta)
        theta1r5 = theta * theta0r5
        theta3 = theta * theta * theta
        theta3r5 = theta3 * theta0r5
        theta4 = theta3 * theta
        theta7r5 = theta4 * theta3r5
        wagner_saturation_pressure_pa = 22064000.0
        if T_K > 0 and T_K < Tc_K:
            # Double precision range in ln --> [-690, 690]
            exponent_factor = (Tc_K / T_K) * (
                -7.85951783 * theta
                + 1.84408259 * theta1r5
                - 11.7866497 * theta3
                + 22.6807411 * theta3r5
                - 15.9618719 * theta4
                + 1.80122502 * theta7r5
            )
            if exponent_factor > 400:
                # T_K \approx 0
                saturation_pressure_pa = 0.0
                rho_kg_per_m3 = dryair_density_kg_per_m3
            elif exponent_factor < -400:
                # T_K \approx Tc_K
                saturation_pressure_pa = wagner_saturation_pressure_pa
                rho_kg_per_m3 = dryair_density_kg_per_m3 * (1.0 - 0.378 * phi * saturation_pressure_pa / pressure_pa)
            else:
                saturation_pressure_pa = wagner_saturation_pressure_pa * np.exp(exponent_factor)
                rho_kg_per_m3 = dryair_density_kg_per_m3 * (1.0 - 0.378 * phi * saturation_pressure_pa / pressure_pa)
        else:
            # Saturation vapor pressure is not addapted for Wagner equation
            raise ValueError(f"Unexpected Tempreture range in Wagner equation, rho_kg_per_m3, : T={T}")
    else:
        raise ValueError(f"Unexpected fluid type density_kg_per_m3: {fluid_type}")
    return np.float64(rho_kg_per_m3)


def viscosity_Pa_s(temprature_celsius: Union[float, np.float64], fluid_type: str) -> np.float64:
    if fluid_type == "air":
        T = temprature_celsius
        eta_kg_m1s1 = 1.487e-6 * ((T + 273.15) ** 1.5) / (T + 390.15)
    else:
        raise ValueError(f"Unexpected fluid type viscosity: {fluid_type}")
    return np.float64(eta_kg_m1s1)


def sound_velocity_m_per_s(
    temprature_celsius: Union[float, np.float64],
    pressure_pa: Union[float, np.float64],
    relative_humidity_percent: Union[float, np.float64],
    fluid_type: str,
) -> np.float64:
    if fluid_type == "air":
        T = temprature_celsius
        H = pressure_pa
        phi_percent = relative_humidity_percent
        # sound velocity in dry air
        c_dry = 20.05 * math.sqrt(temprature_celsius + 273.15)
        # sound velocity in air
        c = c_dry * (density_kg_per_m3(T, H, 0.0, fluid_type)) / (density_kg_per_m3(T, H, phi_percent, fluid_type))
    else:
        raise ValueError(f"Unexpected fluid type mach: {fluid_type}")
    return np.float64(c)


def mach_number(
    velocity_m_per_s: Union[float, np.float64],
    temprature_celsius: Union[float, np.float64],
    pressure_pa: Union[float, np.float64],
    relative_humidity_percent: Union[float, np.float64],
    fluid_type: str,
) -> np.float64:
    if fluid_type == "air":
        # sound velocity in air
        c = sound_velocity_m_per_s(temprature_celsius, pressure_pa, relative_humidity_percent, fluid_type)
        mach = velocity_m_per_s / c
    else:
        raise ValueError(f"Unexpected fluid type mach: {fluid_type}")
    return np.float64(mach)


def reynolds_number(
    velocity_m_per_s: Union[float, np.float64],
    characteristic_length_m: Union[float, np.float64],
    temprature_celsius: Union[float, np.float64],
    pressure_pa: Union[float, np.float64],
    relative_humidity_percent: Union[float, np.float64],
    fluid_type: str,
) -> np.float64:
    if fluid_type == "air":
        w = velocity_m_per_s
        L = characteristic_length_m
        T = temprature_celsius
        H = pressure_pa
        phi_percent = relative_humidity_percent
        # Density for given conditions
        rho = density_kg_per_m3(T, H, phi_percent, fluid_type)
        # Viscosity for given conditions
        eta = viscosity_Pa_s(T, fluid_type)
        # Reynolds number
        Re = w * rho * L / eta
    else:
        raise ValueError(f"Unexpected fluid type reynolds: {fluid_type}")
    return np.float64(Re)


def drag_coefficients_sphere(reynolds_number: Union[float, np.float64], mach_number: Union[float, np.float64], fluid_type: str) -> np.float64:
    Re = reynolds_number
    Ma = mach_number

    if fluid_type == "air":
        if Re < 1e-150:
            raise ValueError(f"Unexpected reynolds_number in drag_coefficients_sphere: Re = {Re}")
        elif Re < 45:
            # gamma: Specific heat ratio of air
            gamma = 1.400
            # Kn: Knudsen number
            Kn = math.sqrt(math.pi * gamma * 0.5) * (Ma / Re)
            # fKn: Cunningham correction factor for Kn << 1
            tmp = -0.55 / Kn

            if tmp < -400:  # avoid underflow
                fKn = 1.0 / (1.0 + Kn * 2.514)
            else:
                fKn = 1.0 / (1.0 + Kn * (2.514 + 0.8 * np.exp(tmp)))

            Cd = (24.0 / Re) * (1.0 + 0.15 * Re**0.687) * fKn
        elif Re <= 200000.0:
            if Ma < 1.5:
                CM = 1.65 + 0.65 * np.tanh(4.0 * Ma - 3.4)
            else:
                CM = 2.18 - 0.13 * np.tanh(0.9 * Ma - 2.7)

            if Ma < 0.8:
                GM = 166.0 * Ma**3 + 3.29 * Ma**2 - 10.9 * Ma + 20.0
            else:
                GM = 5.0 + 40.0 / (Ma**3)

            if Ma < 1.0:
                HM = 0.0239 * Ma**3 + 0.212 * Ma**2 - 0.074 * Ma + 1.0
            else:
                HM = 0.93 + 1.0 / (3.5 + Ma**5)

            Cd = (24.0 / Re) * (1.0 + 0.15 * Re**0.687) * HM + 0.43 * CM / (1.0 + (42500.0 / (Re ** (1.16 * CM))) + GM / (Re**0.5))
        else:
            # Re > 200000 is not defineded in this formula
            raise ValueError(f"Unexpected reynolds_number in drag_coefficients_sphere: Re = {Re}")
    else:
        raise ValueError(f"Unexpected fluid type drag_coefficients_sphere: {fluid_type}")

    return np.float64(Cd)


def skin_friction_coefficients(reynolds_number: Union[float, np.float64], fluid_type: str) -> np.float64:
    Re = reynolds_number

    if fluid_type == "air":
        if Re < 5.3e5:
            Cf = 1.328 / math.sqrt(Re)
        elif Re < 2e8:
            Cf = 0.455 / ((math.log10(Re)) ** 2.58) - 1700.0 / Re
        else:
            Cf = 0.455 / ((math.log10(Re)) ** 2.58)
    else:
        raise ValueError(f"Unexpected fluid type drag_coefficients_sphere: {fluid_type}")

    return np.float64(Cf)


def myplot_in_environment(x, y, xlinlog: str, ylinlog: str, xlabel: str, ylabel: str, fname: str) -> None:
    plt.rcParams["font.family"] = "Times New Roman"
    plt.rcParams["font.size"] = 6
    plt.rcParams["mathtext.fontset"] = "stix"
    plt.rcParams["xtick.direction"] = "in"
    plt.rcParams["ytick.direction"] = "in"

    fig, axs = plt.subplots(1, 1, figsize=(6, 4))

    # graph settings
    axs.plot(x, y, ".-", ms=1, label="y(x)")
    axs.legend(loc=(0.7, 0.7))

    # x-axis
    axs.set_xlabel(xlabel)  # x label
    # axs.set_xlim(0, 10)  # range x=[0, 10]
    # axs.xaxis.set_major_locator(tck.MultipleLocator(1))  # x-axis major locator interval
    # axs.xaxis.set_minor_locator(tck.MultipleLocator(0.5))  # x-axis minor locator interval
    axs.grid(axis="x", which="major", color="#cccccc", linestyle="--")  # show major grid x-axis
    axs.grid(axis="x", which="minor", color="#e7e7e7", linestyle="--")  # show minor grid x-axis
    plt.xscale(xlinlog)
    if xlinlog == "log":
        axs.xaxis.set_major_locator(ticker.LogLocator(base=10))

    # y-axis
    axs.set_ylabel(ylabel)  # y label
    # axs.set_ylim(-3, 3)  # range y=[-3, 3]
    # axs.yaxis.set_major_locator(tck.MultipleLocator(1))  # y-axis major locator interval
    # axs.yaxis.set_minor_locator(tck.MultipleLocator(0.5)) # y-axis minor locator interval
    axs.grid(axis="y", which="major", color="#cccccc", linestyle="--")  # show major grid y-axis
    axs.grid(axis="y", which="minor", color="#e7e7e7", linestyle="--")  # show major grid y-axis
    plt.yscale(ylinlog)
    if ylinlog == "log":
        axs.yaxis.set_major_locator(ticker.LogLocator(base=10))

    # Output image
    # Don't show graph in left-bottom region because Bounding-Box will have negative coordinate
    fig.subplots_adjust(left=0.3, right=0.9, bottom=0.3, top=0.9)
    # plt.show()  # Plot graph with GUI
    fig.savefig(fname, dpi=400, bbox_inches="tight")  # output image with dpi=400


if __name__ == "__main__":
    temprature_celsius = 20.0
    pressure_pa = 101325.0
    relative_humidity_percent = 10.0
    fluid_type = "air"
    velocity_m_per_s = 60.0

    # ------

    mach = mach_number(velocity_m_per_s, temprature_celsius, pressure_pa, relative_humidity_percent, fluid_type)

    drag_coefficients_sphere_vector = np.vectorize(drag_coefficients_sphere)
    x = np.logspace(-1.0, np.log10(2.0 * 1e5), 400, endpoint=False, base=10.0)  # Reynolds number [10^-1, 2*10^5]
    y = drag_coefficients_sphere_vector(x, mach, fluid_type)
    myplot_in_environment(x, y, "log", "log", "Re", "Drag coefficient Cd", "drag_coefficients.png")

    skin_friction_vector = np.vectorize(skin_friction_coefficients)
    x = np.logspace(-1.0, 10.0, 400, base=10.0)  # Reynolds number [10^-1, 10^10]
    y = skin_friction_vector(x, fluid_type)
    myplot_in_environment(x, y, "log", "log", "Re", "skin friction coefficients Cl", "skin_friction_coefficients.png")

    sound_velocity_vector = np.vectorize(sound_velocity_m_per_s)
    x = np.linspace(-50, 150, 300)  # temperature
    y = sound_velocity_vector(x, pressure_pa, relative_humidity_percent, fluid_type)
    myplot_in_environment(x, y, "linear", "linear", "Temperture Celsius", "Sound velocity m/s", "sound_velocity.png")

    density_vector = np.vectorize(density_kg_per_m3)
    x = np.linspace(-50, 150, 300)  # temperature
    y = density_vector(x, pressure_pa, relative_humidity_percent, fluid_type)
    myplot_in_environment(x, y, "linear", "linear", "Temperture Celsius", "Density kg/m3", "density.png")

    viscosity_vector = np.vectorize(viscosity_Pa_s)
    x = np.linspace(-50, 150, 300)  # temperature
    y = viscosity_vector(x, fluid_type)
    myplot_in_environment(x, y, "linear", "linear", "Temperture Celsius", "Viscosity Pa*s", "viscosity.png")
